/*
# X-BASED DINO
#
#  Dino2d.h
#
###
#
#  Copyright (c) 1995	David Albert Bagley, bagleyd@source.asset.com
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Public header file for Dino2d */

#ifndef _XtDino2d_h
#define _XtDino2d_h

/***********************************************************************
 *
 * Dino2d Widget
 *
 ***********************************************************************/

#include "Dino.h"

typedef struct _Dino2DClassRec *Dino2DWidgetClass;
typedef struct _Dino2DRec *Dino2DWidget;

extern WidgetClass dino2dWidgetClass;

#endif _XtDino2d_h
/* DON'T ADD STUFF AFTER THIS #endif */
